from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Tuple, Union

import requests

from .boxes import PartialTag, ResultList, Tag, _parse_results
from .pages import FrontPage, Member
from .utils import BASE_URL, generate_login_cookies, get_page, get_page_type, request, soup

if TYPE_CHECKING:
    from .enums import RSSType, SearchCategory


__all__ = ["search", "parse_page", "login", "logout", "front_page", "parse_results"]


def search(
    search_category: SearchCategory,
    *,
    query: str = None,
    sort: Tuple[str, str] = None,
    page: int = 1,
    **filters,
) -> ResultList:
    """Search for for a certain type of models and return a list of thumbnails of that
    model. This function was created to make full use of moddb's filter and sorting system as
    long as the appropriate parameters are passed. Because this function is a single one for every
    model type in moddb the parameters that can be passed vary between model type but the
    documentation will do its best to document all the possibilities. All the objects listed in this
    function are enumerations. Finally it is key to note that the result attribute of the returned search
    object is not a list of the parsed pages but a list of Thumbnail objects containing the image, url,
    name and type of the object on which. Call the thumbnail's parse method and it will return the full model.

    Parameters
    ------------
    search_category : SearchCategory
        The model type that you want to search
    query : str
        String to search for in the model title
    sort : Tuple[str, str]
        The tuple to sort by, look at the models your are searching for documentation
        on sorting.
    page : int
        The page of results to get first
    filters : dict
        Search filters

    Returns
    --------
    ResultList
        The search object containing the current query settings (so at to be able to redo the search easily),
        pagination metadata and helper methods to navigate the list of results.
    """
    sort_ready = f"{sort[0]}-{sort[1]}" if sort else None

    game = filters.get("game", None)
    game = game.id if game else None

    url = f"{BASE_URL}/{search_category.name}/page/{page}"
    filter_parsed = {key: value.value for key, value in filters.items() if hasattr(value, "value")}

    params = {
        "filter": "t",
        "kw": query,
        "sort": sort_ready,
        "game": game,
        "year": filters.get("years", None),
        **filter_parsed,
    }

    html = get_page(
        url,
        params=params,
    )

    results, current_page, total_pages, total_results = _parse_results(html)

    return ResultList(
        results=results,
        total_pages=total_pages,
        current_page=current_page,
        params=params,
        url=f"{BASE_URL}/{search_category.name}",
        total_results=total_results,
    )


def parse_page(url: str) -> Any:
    """Parse a url and return the appropriate object.

    Parameters
    ------------
    url : str
        The url to parse

    Returns
    --------
    Model
        The parsed page as the instance of the model the page represents, can be anything like
        Mod or Game
    """

    html = get_page(url)
    page_type = get_page_type(url)

    model = getattr(sys.modules["moddb"], page_type.name.title())(html)
    return model


def parse_results(url: str, *, params: dict = {}) -> ResultList:
    """Parse a list of results and return them as a
    list of thumbnails.

    Parameters
    -----------
    url : str
        The url of the result list to parse

    Returns
    --------
    ResultList
        The list of thumbnails, wrapped in a ResultList
        so as to benefit from the helper methods that
        help with navigation
    """

    resp = request(requests.Request("GET", url, params=params))
    html = soup(resp.text)

    results, current_page, total_pages, total_results = _parse_results(html)

    parts = resp.url.split("?")
    url = parts[0].split("/page")[0]

    filters = {}
    if len(parts) > 1:
        for filter in parts[1].split("&"):
            key, value = filter.split("=")
            filters[key] = value

    return ResultList(
        results=results,
        total_pages=total_pages,
        current_page=current_page,
        url=url,
        total_results=total_results,
        params=filters,
    )


def login(username: str, password: str) -> Member:
    """Login the user to moddb through the library, this allows user to see guest comments and see
    private groups they are part of.

    Parameters
    -----------
    username : str
        The username of the user
    password : str
        The password associated to that username

    Raises
    -------
    ValueError
        The password or username was incorrect

    Returns
    --------
    Member
        The member you are logged in as
    """

    sys.modules["moddb"].SESSION.cookies = generate_login_cookies(username, password)
    return Member(get_page(f"{BASE_URL}/members/{username.replace('_', '-')}"))


def logout():
    """Logs the user out by clearing the cookies, all unapproved guest comments will be hidden and
    all private groups will be hidden once more
    """
    sys.modules["moddb"].SESSION.cookies.clear()


def front_page() -> FrontPage:
    """This returns a model representing the front page of  May sound fancy but it is no more
    than a collection of popular mods, files, games and articles. In addition jobs are listed
    and a poll.

    Returns
    --------
    FrontPage
        The front page object.

    """
    html = get_page(BASE_URL)
    return FrontPage(html)


def rss(rss_type: RSSType):
    """Get the RSS feed url for the entire site depending on which feed type you want

    Parameters
    -----------
    rss_type : RSSType
        The type of feed you desire to get

    Returns
    --------
    str
        URL for the feed type
    """
    return f"https://rss.moddb.com/{rss_type.name}/feed/rss.xml"


def search_tags(tag: Union[str, PartialTag, Tag]):
    """Search for entities tagged with a tag-id or Tag.

    Parameters
    -----------
    tag : Union[str, PartialTag, Tag]
        Either a name-id or the Tag to search for

    Returns
    --------
    ResultList[Thumbnail]
        List of entities tagged with this
    """
    if isinstance(tag, (Tag, PartialTag)):
        tag = tag.name_id

    url = f"https://www.moddb.com/tags/{tag}"
    return parse_results(url)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Tuple

from ..boxes import (
    MemberProfile,
    MemberStatistics,
    PartialTag,
    Profile,
    ResultList,
    Statistics,
    Thumbnail,
    _parse_results,
)
from ..enums import GroupCategory, Membership, SearchCategory, ThumbnailType, TimeFrame
from ..utils import LOGGER, concat_docs, get_page, join
from .article import Blog
from .base import BaseMetaClass, PageMetaClass
from .mixins import GetAddonsMixin, GetEnginesMixin, GetGamesMixin, GetModsMixin, GetWaresMixin

if TYPE_CHECKING:
    import bs4

    from ..boxes import CommentList


@concat_docs
class Group(PageMetaClass, GetAddonsMixin):
    """This object represents the group model of  Certain attributes can be None if the group
    has been set to private. If you wish to see a group you have access to then you can login with the
    login

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    subscriptions : :class:`.Membership`
        The subscription system of the group (private, invitation)
    category : :class:`.GroupCategory`
        The category of the group (funny, literature)

    Sorting
    --------
        * **id** - order group by date, asc is most recent first, desc is oldest first
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **membercount** - order by number of members, asc is most members first, desc is lest member first

    Attributes
    -----------
    name : str
        The name of the group
    private : bool
        Whether or not the group is private
    profile : Profile
        The profile object for the group
    stats : Statistics
        The stats of the Group
    tags : List[PartialTag]
        A list of partial tags. You can use `get_tags` and then use the name id to get the right one.
    embed : str
        The html for athe group embed
    medias : List[Thumbnail]
        A list of media like thumbnail objects representing all the images, videos and audio
        clips that a group has published.
    suggestions : List[Thumbnail]
        A list of group like thumbnail objects representing the suggestions made by moddb to
        members visiting this group
    articles : List[Thumbnail]
        A list of article like thumbnail objects representing some of the articles published
        by the group
    description : str
        The plaintext description of the group
    """

    get_reviews = None

    def __init__(self, html: bs4.BeautifulSoup):
        self.name = html.find("div", class_="title").h2.a.string
        BaseMetaClass.__init__(self, html)
        self.private = False

        try:
            self.profile = Profile(html)
        except AttributeError:
            LOGGER.info(
                "Entity '%s' has no profile (private)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.profile = None
            self.private = True

        try:
            self.stats = Statistics(html)
        except AttributeError:
            LOGGER.info(
                "Entity '%s' has no stats (private)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.stats = None

        try:
            raw_tags = html.find("form", attrs={"name": "tagsform"}).find_all("a")
            self.tags = [
                PartialTag(x.string, join(x["href"]), x["href"].split("/")[-1])
                for x in raw_tags
                if x.string is not None
            ]
        except AttributeError:
            LOGGER.info(
                "Entity '%s' has no tags (private)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.tags = []

        try:
            self.embed = html.find("input", type="text", class_="text textembed")["value"]
        except TypeError:
            try:
                self.embed = str(html.find_all("textarea")[1].a)
            except IndexError:
                LOGGER.info(
                    "Group '%s' has no embed", self.name, exc_info=LOGGER.level >= logging.DEBUG
                )
                self.embed = None

        self.suggestions = self._get_suggestions(html)

        try:
            articles_raw = html.find("span", string="Articles").parent.parent.parent.find(
                "div", class_="table"
            )
            thumbnails = articles_raw.find_all(
                "div", class_="row rowcontent clear", recursive=False
            )
            self.articles = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    image=x.a.img["src"],
                    summary=x.find("p").string,
                    date=x.find("time")["datetime"],
                    type=ThumbnailType.article,
                )
                for x in thumbnails
            ]
        except AttributeError:
            LOGGER.info(
                "Group '%s' has no article suggestions",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.articles = []

        try:
            self.description = html.find("div", id="profiledescription").text
        except AttributeError:
            self.description = (
                html.find("div", class_=["column", "span-all"])
                .find("div", class_="tooltip")
                .parent.text
            )

        self.medias = self._get_media(2, html=html)

    def _get_suggestions(self, html: bs4.BeautifulSoup) -> List[Thumbnail]:
        """Hidden method used to get the list of suggestions on the page. As with most things, this list of suggestions
        will be a list of Thumbnail objects that can be parsed individually. Slightly modified to fit a group

        Parameters
        -----------
        html : bs4.BeautifulSoup
            The html page we are trying to parse the suggestions for

        Returns
        --------
        List[Thumbnail]
            The list of suggestions as thumbnails.
        """
        try:
            suggestions_raw = (
                html.find("span", string=("You may also like", "Popular Articles"))
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)
            )
        except AttributeError:
            LOGGER.info(
                "Group '%s' has no sidebar suggestions",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            return []

        suggestions = []
        for x in suggestions_raw:
            try:
                link = x.find("a", class_="image")
                suggestion_type = link["href"].split("/")[1].replace("s", "")
                suggestion = Thumbnail(
                    name=link["title"],
                    url=link["href"],
                    image=link.img["src"],
                    type=ThumbnailType[suggestion_type],
                )
                suggestions.append(suggestion)
            except (AttributeError, TypeError):
                pass

        return suggestions

    def __repr__(self):
        return f"<Group name={self.name}>"


@concat_docs
class Team(Group, GetEnginesMixin, GetGamesMixin, GetModsMixin, GetWaresMixin):
    """A team is a group of people, which are the author of a game, a mod or an engine. A group has members which all
    have rights on those page. Like a member but instead of a single person authoring various mods and games it's several.

    Parameters
    ------------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    subscriptions : :class:`.Membership`
        The subscription system of the company (private, invitation)
    category : :class:`.TeamCategory`
        What does the team do (publisher, developer)

    Sorting
    --------
        * **id** - order by creation, desc is most recent first, asc is oldest first
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **game** - order by game???
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first

    Attributes
    -----------
    games : List[Thumbnail]
        A list of game like thumbnails that the team has authored.
    engines : List[Thumbnail]
        A list of engine like objects that the team has authored.

    """

    def __init__(self, html: bs4.BeautifulSoup):
        super().__init__(html)
        try:
            self.games = self._get_games(html)
        except AttributeError:
            LOGGER.info("Team '%s' has no games", self.name, exc_info=LOGGER.level >= logging.DEBUG)
            self.games = []

        try:
            self.engines = self._get_engines(html)
        except AttributeError:
            LOGGER.info(
                "Team '%s' has no engines", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.engines = []
        try:
            mods = (
                html.find("span", string="Popular Mods")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)[1:]
            )
            self.mods = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    type=ThumbnailType.mod,
                    image=x.a.img["src"],
                )
                for x in mods
            ]
        except AttributeError:
            LOGGER.info("Team '%s' has no mods", self.name, exc_info=LOGGER.level >= logging.DEBUG)
            self.mods = []


@concat_docs
class Member(PageMetaClass, GetGamesMixin, GetModsMixin, GetAddonsMixin):
    """The object to represent an individual member on ModDB

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Sorting
    --------
    * **username** - sort alphabetically by username, asc is z-a and desc is a-z
    * **id** - sort by member creation date, asc is most recent, desc is oldest
    * **online** - sort by last online, asc is most recently seen online and desc is least recently

    Attributes
    -----------
    profile : MemberProfile
        Since member profile boxes have no overlap with other profiles, they are a separate object type
        but serve the same function of making the side box "Profile" into an object, except exclusively
        for a member page.
    stats : MemberStatistics
        Since member statistics have no overlap with regular statistic pages, they are a separate object type
        but serve the same function of making the side box "Statistics" into an object, but exclusively for
        the member page.
    description : str
        Description written on the member profile
    groups : List[Thumbnail]
        A list of group/team like thumbnail objects representing both the Teams the member is part of and the
        Groups the member is part of.
    blog : Blog
        The front page blog shown on the member page
    blogs : List[Thumbnail]
        A list of blog like thumbnails representing the blog suggestion of a member's frontpage
    friends : List[Thumnails]
        A list of member like thumbnails representing some of the friends shown on the member's front
        page
    """

    def __init__(self, html: bs4.BeautifulSoup):
        super().__init__(html, SearchCategory.members)
        try:
            self.profile = MemberProfile(html)
        except AttributeError:
            LOGGER.info(
                "Member '%s' has no profile (private)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.profile = None

        try:
            self.stats = MemberStatistics(html)
        except AttributeError:
            LOGGER.info(
                "Member '%s' has no stats (private)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.stats = None

        try:
            self.description = html.find("div", id="profiledescription").p.string
        except AttributeError:
            LOGGER.info(
                "Member '%s' has no description", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.description = None

        try:
            groups_raw = (
                html.find("span", string="Groups")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)[:-2]
            )
            self.groups = [
                Thumbnail(name=div.a["title"], url=div.a["href"], type=ThumbnailType.group)
                for div in groups_raw
            ]
        except AttributeError:
            LOGGER.info(
                "Member '%s' doesn't have any groups",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.groups = []

        try:
            blogs_raw = (
                html.find("span", string="My Blogs")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)
            )
            self.blog = Blog(heading=blogs_raw.pop(0), text=blogs_raw.pop(0))
        except (TypeError, AttributeError):
            self.blog = None
            LOGGER.info(
                "Member '%s' has no front page blog",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            blogs_raw = (
                html.find("span", string="My Blogs")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)
            )
            self.blogs = [
                Thumbnail(name=blog.a.string, url=blog.a["href"], type=ThumbnailType.blog)
                for blog in blogs_raw[:-2]
            ]
        except (TypeError, AttributeError):
            self.blogs = []
            LOGGER.info(
                "Member '%s' has no blog suggestions",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            friends = html.find("div", class_="table tablerelated").find_all(
                "div", recursive=False
            )[1:]
            self.friends = [
                Thumbnail(
                    name=friend.a["title"],
                    url=friend.a["href"],
                    type=ThumbnailType.member,
                )
                for friend in friends
            ]
        except AttributeError:
            self.friends = []
            LOGGER.info(
                "Member '%s' has no friends ;(", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )

    def __repr__(self):
        return f"<Member name={self.name} level={self.profile.level}>"

    def get_blogs(
        self,
        index: int = 1,
        *,
        query: str = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Search through a member's blogs one page at a time with certain filters.

        Parameters
        -----------
        index : int
            The page index you wish to get the blogs for, allows to hop around.
        timeframe : TimeFrame
            The date the blog was added, optional
        query : str
            The string to look for in the blog title, optional.
        sort : Tuple[str, str]
            The sorting tuple to sort by

        Returns
        --------
        ResultList[Blog]
            The list of blogs on this page.
        """
        params = {
            "filter": "t",
            "kw": query,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        url = f"{self.url}/blogs"
        html = get_page(f"{url}/page/{index}", params=params)
        results, current_page, total_pages, total_results = _parse_results(html)

        return ResultList(
            results=results,
            params=params,
            url=url,
            current_page=current_page,
            total_pages=total_pages,
            total_results=total_results,
        )

    def get_member_comments(self, index: int = 1, *, show_deleted: bool = False) -> CommentList:
        """Gets a page of all the comments a member has posted.

        Parameters
        -----------
        index : int
            The page number to get the comments from.
        show_deleted : Optional[bool]
            Pass true to show deleted user comments. Only works if it is a page
            you have permissions on.

        Returns
        --------
        CommentList[Comment]
            A list of the comments made by the user.

        """
        params = {"deleted": "t" if show_deleted else None}

        html = get_page(f"{self.url}/comments/page/{index}", params=params)
        return self._get_comments(html)

    def get_friends(self, index: int = 1, *, username: str = None) -> ResultList:
        """Get a page of the friends of the member

        Parameters
        -----------
        index : int
            The page number to get the friends from.
        username : Optional[str]
            The username of the user you are looking for

        Returns
        --------
        ResultList[Thumbnail]
            A list of member like thumbnails of the member's friends
        """
        params = {"filter": "t", "username": username}

        return self._get(f"{self.url}/friends/page/{index}", params=params)

    def get_groups(
        self,
        index: int = 1,
        *,
        query: str = None,
        subscription: Membership = None,
        category: GroupCategory = None,
    ) -> ResultList:
        """Get a page of the groups and teams a member is part of.

        Parameters
        -----------
        index : int
            The page number to get the friends from.
        query : Optional[str]
            The text to look for in the group's name
        subscription : Optional[Membership]
            The membership rules
        category : Optional[GroupCategory]
            The category of groups to search for

        Returns
        --------
        ResultList[Thumbnail]
            A list of team/group like thumbnails the member is part of
        """
        params = {
            "filter": "t",
            "kw": query,
            "subscription": subscription.value if subscription else None,
            "category": category.value if category else None,
        }

        return self._get(f"{self.url}/groups/page/{index}", params=params)



from __future__ import annotations

import collections
import datetime
import logging
import re
import sys
from typing import TYPE_CHECKING, Any, Generic, List, Tuple, TypeVar

from typing_extensions import Self
from bs4 import BeautifulSoup

from .enums import (
    AddonCategory,
    ArticleCategory,
    Genre,
    GroupCategory,
    HardwareCategory,
    Licence,
    Membership,
    PlayerStyle,
    Scope,
    SearchCategory,
    SoftwareCategory,
    Status,
    TeamCategory,
    Theme,
    ThumbnailType,
)
from .utils import (
    BASE_URL,
    LOGGER,
    generate_hash,
    get,
    get_date,
    get_list_stats,
    get_page,
    get_page_type,
    get_siteareaid,
    get_views,
    join,
    normalize,
    time_mapping,
    unroll_number,
)

if TYPE_CHECKING:
    from .pages.article import Article

__all__ = [
    "Statistics",
    "Profile",
    "Style",
    "Thumbnail",
    "Comment",
    "MissingComment",
    "MemberProfile",
    "MemberStatistics",
    "PlatformStatistics",
    "PartialArticle",
    "Option",
    "Mirror",
    "ResultList",
    "CommentList",
]


class Statistics:
    """The stats box, on pages that have one. This represents total stats and daily stats in one
    neat package.

    Attributes
    ----------
    files : int
        The number of files this page has uploaded
    articles : int
        The number of articles this page has uploaded
    reviews : int
        The number of reviews this page has been given
    watchers : int
        The number of people following this page
    mods : int
        The number of mods this page is related too (only applies to games, members and teams)
    addons : int
        The number of addons this page has uploaded
    members : int
        The number of members a group has (only applies to groups and teams)
    visits : int
        The total number of times this page has been viewed
    today : int
        The number of times this page has been viewed today
    rank : int
        The current rank of the page against all other pages of the same type
    total : int
        The maximum rank number
    updated : datetime.datetime
        The last time this page was updated
    """

    def __init__(self, html: BeautifulSoup):
        misc = html.find_all(
            "h5",
            string=(
                "Files",
                "Articles",
                "Reviews",
                "Watchers",
                "Mods",
                "Addons",
                "Members",
            ),
        )
        self.__dict__.update(
            {stat.string.lower(): int(normalize(stat.parent.a.string)) for stat in misc}
        )

        visits = normalize(html.find("h5", string="Visits").parent.a.string)
        self.visits, self.today = get_views(visits)

        rank = normalize(html.find("h5", string="Rank").parent.a.string).split("of")
        self.rank = int(rank[0].replace(",", ""))
        self.total = int(rank[1].replace(",", ""))

        try:
            self.updated = get_date(html.find("time", itemprop="dateModified")["datetime"])
        except TypeError:
            self.updated = None

    def __repr__(self):
        return f"<Statistics rank={self.rank}/{self.total}>"


class Profile:
    """The profile object is used for several models and as such attribute vary based on which model
    the profile is attached too. Profiles are only present on Mod, Game, Member, Addon, Engine, Company,
    Hardware, Software and Group pages.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    category : Union[AddonCategory, HardwareCategory, SoftwareCategory, TeamCategory, GroupCategory, SearchCategory]
        The category the page falls under within the context of what the page is. E.g the page is an Addon category
        will be an AddonCategory enum. If the category of the page doesn't fall under any of the above mentionned
        the attribute will be of type SearchCategory.
    contact : str
        The url to contact the page owner
    follow : str
        The url to click to follow the mod
    share : dict
        A dictionnary of share links with the place they will be shared as the key and the url
        for sharing as the value.
    private : bool
        Exclusive to Group and Team, True if the group is private, else False
    membership : Membership
        Exclusive to Group and Team, represents the join procedure (invitation only, private, public)
    icon : str
        Exclusive to Game, Mod and Addon pages. URL of the icon image
    developers : dict
        Exclusive to Game, Mods, Engine and Addon pages. Dictionnary of member/team like thumbnails as
        values and the role of the member/team as the key (creator, publisher, developer, ect...)
    release : datetime.datetime
        Exclusive to Game, Mods, Engine and Addon pages. Datetime object of when the page was
        released, can be None if the page hasn't seen a release yet.
    homepage : str
        Present on all pages but Group pages. URL to the page's homepage. Can be None
    engine : Thumbnail
        Exclusive to Game and Addon pages. Engine like thumbnails representing the engine the addon/game
        was built for.
    game : Thumbnail
        Exclusive to Mod pages. Game like thumbnail representing the game the mod was built for.
    licence : Licence
        Exclusive to Engine and Addon pages. Object representing the licence the engine operates under.
    platforms : List[Thumbnail]
        Exclusive to Game, Engine and Addon pages. List of platform like thumbnails representing
        the plaftorms the software was built for.
    status : Status
        Exclusive to Games, Mods, Addons, Engines, Hardware .Whether the thing is released, unreleased, ect...
    download_count: int
        Total count of all downloads on the page, this adds up downloads of all files and addons. Exclusive
        to mods and games.

    """

    def __init__(self, html: BeautifulSoup):
        try:
            _name = html.find("a", itemprop="mainEntityOfPage").string
        except AttributeError:
            try:
                _name = html.find("span", itemprop="headline").string
            except AttributeError:
                _name = html.find("div", class_="title").h2.a.string
        try:
            url = html.find("meta", property="og:url")["content"]
        except TypeError:
            url = join(html.find("a", string=self.name)["href"])

        regex = r"\/([a-z]+)\/"
        matches = re.findall(regex, url)
        matches.reverse()
        page_type = SearchCategory[matches[0] if matches[0].endswith("s") else matches[0] + "s"]

        self.category = page_type
        profile_raw = html.find("span", string="Profile").parent.parent.parent.find(
            "div", class_="table tablemenu"
        )
        self.contact = join(html.find("h5", string="Contact").parent.span.a["href"])

        self.follow = join(html.find("a", title="Follow")["href"])

        try:
            share = profile_raw.find("h5", string="Share").parent.span.find_all("a")
            self.share = {
                "reddit": share[0]["href"],
                "mail": share[1]["href"],
                "twitter": share[2]["href"],
                "facebook": share[3]["href"],
            }
        except (AttributeError, IndexError):
            LOGGER.info(
                "Something funky about share box of %s %s",
                page_type.name,
                _name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.share = None

        if page_type in [SearchCategory.developers, SearchCategory.groups]:
            self.private = (
                profile_raw.find("h5", string="Privacy").parent.span.string.strip() != "Public"
            )

            membership = profile_raw.find("h5", string="Subscription").parent.span.string.strip()
            if membership == "Open to all members":
                self.membership = Membership(3)
            elif membership == "Must apply to join":
                self.membership = Membership(2)
            else:
                self.membership = Membership(1)

        if page_type in [
            SearchCategory.games,
            SearchCategory.mods,
            SearchCategory.addons,
        ]:
            try:
                self.icon = profile_raw.find("h5", string="Icon").parent.span.img["src"]
            except AttributeError:
                self.icon = None
                LOGGER.info(
                    "%s '%s' does not have an icon",
                    page_type,
                    _name,
                    exc_info=LOGGER.level >= logging.DEBUG,
                )

        if page_type in [
            SearchCategory.games,
            SearchCategory.mods,
            SearchCategory.engines,
            SearchCategory.addons,
            SearchCategory.hardwares,
            SearchCategory.softwares,
        ]:
            people = profile_raw.find_all(
                "h5",
                string=[
                    "Developer",
                    "Publisher",
                    "Developer & Publisher",
                    "Creator",
                    "Company",
                ],
            )
            self.developers = {
                x.string.lower(): Thumbnail(
                    url=x.parent.a["href"],
                    name=x.parent.a.string,
                    type=ThumbnailType.team if x.string != "Creator" else ThumbnailType.member,
                )
                for x in people
            }

            try:
                d = profile_raw.find("h5", string="Release date").parent.span.time
                self.release = get_date(d["datetime"])
            except KeyError:
                LOGGER.info(
                    "%s %s has not been released",
                    page_type.name,
                    _name,
                    exc_info=LOGGER.level >= logging.DEBUG,
                )
                self.release = None

            if "Coming" in d.string:
                self.status = Status.coming_soon
            elif "Early" in d.string:
                self.status = Status.early_access
            elif "Released" in d.string:
                self.status = Status.released
            else:
                self.status = Status.unreleased

            if page_type != SearchCategory.mods:
                platforms = profile_raw.find("h5", string="Platforms").parent.span.find_all("a")
                self.platforms = [
                    Thumbnail(name=x.string, url=x["href"], type=ThumbnailType.platform)
                    for x in platforms
                ]

        if page_type != SearchCategory.groups:
            try:
                self.homepage = html.find("h5", string="Homepage").parent.span.a["href"]
            except AttributeError:
                self.homepage = None
                LOGGER.info(
                    "%s %s has no homepage",
                    page_type.name,
                    _name,
                    exc_info=LOGGER.level >= logging.DEBUG,
                )

        if page_type in [SearchCategory.games, SearchCategory.addons]:
            engine = profile_raw.find("h5", string="Engine")
            url = engine.parent.span.a["href"]
            name = engine.parent.span.a.string
            self.engine = Thumbnail(url=url, name=name, type=ThumbnailType.engine)

        if page_type == SearchCategory.mods:
            game = profile_raw.find("h5", string="Game")
            url = game.parent.span.a["href"]
            name = game.parent.span.a.string
            self.game = Thumbnail(url=url, name=name, type=ThumbnailType.game)

        if page_type in [SearchCategory.engines, SearchCategory.addons]:
            self.licence = Licence(
                int(profile_raw.find("h5", string="Licence").parent.span.a["href"].split("=")[-1])
            )

        if page_type == SearchCategory.hardwares:
            self.category = HardwareCategory(
                int(profile_raw.find("h5", string="Category").parent.span.a["href"].split("=")[-1])
            )

        if page_type == SearchCategory.softwares:
            self.category = SoftwareCategory(
                int(profile_raw.find("h5", string="Category").parent.span.a["href"].split("=")[-1])
            )

        if page_type == SearchCategory.addons:
            self.category = AddonCategory(
                int(profile_raw.find("h5", string="Category").parent.span.a["href"].split("=")[-1])
            )

        if page_type == SearchCategory.developers:
            category = html.find("h3").string.strip().lower()
            try:
                self.category = TeamCategory[category]
            except KeyError:
                self.category = TeamCategory(7)

        if page_type == SearchCategory.groups:
            category = html.find("h3").string.strip().lower().replace(" & ", "_")
            self.category = GroupCategory[category]

        if page_type in [SearchCategory.games, SearchCategory.mods]:
            try:
                self.download_count = unroll_number(
                    html.find("a", class_="downloadautotoggle").span.string
                )
            except AttributeError:
                self.download_count = 0
                LOGGER.info(
                    "%s %s has no download count",
                    page_type.name,
                    _name,
                    exc_info=LOGGER.level >= logging.DEBUG,
                )

    def __repr__(self):
        return f"<Profile category={self.category.name}>"


class Style:
    """Represents semantic information on the page's theme.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    ----------
    theme : Theme
        fantasy, sci-fi, ect...
    genre : Genre
        fps, rpg, moba, ect...
    players : PlayerStyle
        Singplayer, multiplayer, ect...
    scope : Scope
        Triple A games or indie
    boxart : str
        URL of the boxart for the page.
    """

    def __init__(self, html: BeautifulSoup):
        misc = html.find_all("h5", string=("Theme", "Genre", "Players"))
        styles = {
            style.string.lower(): re.findall(r"(\d*)$", style.parent.a["href"])[0] for style in misc
        }

        self.theme = Theme(int(styles["theme"]))
        self.genre = Genre(int(styles["genre"]))
        self.players = PlayerStyle(int(styles["theme"]))

        try:
            self.scope = Scope(int(html.find("h5", string="Project").parent.a["href"][-1]))
        except AttributeError:
            LOGGER.info("Has no scope", exc_info=LOGGER.level >= logging.DEBUG)

        try:
            self.boxart = html.find("h5", string="Boxart").parent.span.a.img["src"]
        except AttributeError:
            LOGGER.info("Has no boxart", exc_info=LOGGER.level >= logging.DEBUG)

    def __repr__(self):
        return (
            f"<Style genre={self.genre.name} theme={self.theme.name} players={str(self.players)}>"
        )


class Thumbnail:
    """Thumbnail objects are minature version of ModDB models. They can be parsed to return the full
    version of the model.

    Attributes
    -----------
    url : str
        The url to the full model, mandatory attribute.
    name : str
        The name of the model
    image : str
        The optional thumbnail image of the model
    summary : str
        Optional bit of fluff
    date : datetime.datetime
        A date related to this timestamp if it exists. Can be None
    type : ThumbnailType
        The type of the resource, mandatory attribute

    """

    def __init__(self, **attrs):
        self.url: str = join(attrs.get("url"))
        self.name: str | None = attrs.get("name", None)
        self.image: str | None = attrs.get("image", None)
        self.summary: str | None = attrs.get("summary", None)
        self.date: datetime.datetime | None = attrs.get("date", None)
        self.type: ThumbnailType = attrs.get("type")

    def __repr__(self):
        return f"<Thumbnail name={self.name} type={self.type.name}>"

    def parse(self) -> Any:
        """Uses the Thumbnail's mandatory attributes to get the full html of the
        model and parse them with the appropriate object.

        Returns
        --------
        Any
            The model that was parsed, can be any model from the list of the ThumbnailType
            enum.
        """
        return getattr(sys.modules["moddb"], self.type.name.title())(get_page(self.url))


def _parse_results(html):
    result_box = html.find("div", class_="normalbox browsebox")
    try:
        search_raws = (
            result_box.find("div", class_="inner")
            .find("div", class_="table")
            .find_all("div", class_=["rowcontent"])
        )
    except AttributeError:
        return [], 1, 1, 0

    results = []
    try:
        for obj in search_raws:
            date = obj.find("time")
            summary = obj.find("p")
            results.append(
                Thumbnail(
                    name=obj.a["title"],
                    url=obj.a["href"],
                    image=obj.a.img["src"],
                    type=get_page_type(join(obj.a["href"])),
                    summary=summary.string if summary else None,
                    date=get_date(date["datetime"]) if date and date.string != "TBD" else None,
                )
            )
    except (TypeError, KeyError):
        # parse as a title-content pair of articles
        LOGGER.info(
            "Parsing articles as key-value pair list", exc_info=LOGGER.level >= logging.DEBUG
        )
        for title, content in zip(search_raws[::2], search_raws[1::2]):
            date = title.find("time")
            url = title.find("h4").a
            results.append(
                Thumbnail(
                    name=url.text,
                    url=url["href"],
                    image=None,
                    type=get_page_type(join(url["href"])),
                    summary=content.text,
                    date=get_date(date["datetime"]) if date and date.string != "TBD" else None,
                )
            )

    current_page, total_page, total_results = get_list_stats(result_box)
    if total_results is None:
        total_results = len(results)

    return results, current_page, total_page, total_results


def _parse_comments(html):
    comments = []
    comment_box = html.find("div", id="comments")
    if comment_box is None:
        return [], 1, 1, 0

    current_page, total_page, total_results = get_list_stats(comment_box)

    try:
        url = html.find("meta", property="og:url")["content"]
    except TypeError:
        url = join(html.find("a", itemprop="mainEntityOfPage")["href"])

    comments_raw = comment_box.find("div", class_=["tablecomments"]).find_all(
        "div", class_="row", recursive=False
    )
    if total_results is None:
        total_results = len(comments_raw)

    for raw in comments_raw:
        comment = Comment(raw)
        comment._url = f"{url}/page/{current_page}"
        if comment.position == 1:
            try:
                comments[-1].children.append(comment)
            except IndexError:
                comments.append(MissingComment(0))
                comments[-1].children.append(comment)
        elif comment.position == 2:
            try:
                comments[-1].children[-1].children.append(comment)
            except IndexError:
                try:
                    comments[-1].children.append(MissingComment(1))
                    comments[-1].children[-1].children.append(comment)
                except IndexError:
                    comments.append(MissingComment(0))
                    comments[-1].children.append(MissingComment(1))
                    comments[-1].children[-1].children.append(comment)
        else:
            comments.append(comment)

    return comments, current_page, total_page, total_results


class CommentAuthor(Thumbnail):
    """Represents the thumbnail of a user having left a comment on a page. Functions the same as a
    thumbnail but with an extra attribute.

    Attributes
    -----------
    comment_count : int
        Number of comments the user has posted
    """

    def __init__(self, **attrs):
        super().__init__(**attrs)

        self.comment_count: int = attrs.get("comment_count", 0)

    def __repr__(self):
        return (
            f"<Thumbnail name={self.name} type={self.type.name} comment_count={self.comment_count}>"
        )


class Comment:
    """A moddb comment object.

    Parameters
    -----------
    html : bs4.Tag
        The html to parse into the object. Must be the exact div of the comment.

    Attributes
    -----------
    id : int
        The ID of the comment
    author : Thumbnail
        A member like thumbnail of the member who posted the comment
    date : datetime.datetime
        Date and time of the comment creation
    position : int
        Ranging from 0-2 represents the nested level of the comment.
    children : int
        Comment object replying directly to this one. If the comment is
        parsed on its own it will be null. It is only populated if originating
        from a CommentList
    content : str
        Text of the comment can be none if the comment only contains embeds
    embeds : list
        List of urls that have been embeded
    karma : int
        The current karma count
    upvote : str
        Link to upvote the comment
    downvote : str
        Link to downvote the comment
    approved : bool
        Whether or not the comment is still waiting for admin approval and is visible to the guest members
    developer : bool
        Whether or not the comment was posted one of the page creators
    staff : bool
        Wether or not the comment was posted by one of moddb's staff members
    subscriber : bool
        Whether or not the comment was posted by a moddb subscriber
    guest : bool
        Whether or not the comment was posted by a guest user
    location : Thumbnail
        Thumbnail of the place the comment is, only available when getting comments from get_member_comments. This
        thumbnail does not guarantee that you will find the comment if you parse it, since the url does not
        contain the page number.
    """

    def __init__(self, html: BeautifulSoup):
        author = html.find("a", class_="avatar")
        self.id = int(html["id"])
        comment_count = int(
            html.find("span", class_="heading")
            .text.strip()
            .split("-")[-1]
            .replace("comments", "")
            .replace(",", "")
        )
        self.author = CommentAuthor(
            name=author["title"],
            url=author["href"],
            image=author.img["src"],
            type=ThumbnailType.member,
            comment_count=comment_count,
        )
        self.date = get_date(html.find("time")["datetime"])
        actions = html.find("span", class_="actions")
        self._fetch_time = datetime.datetime.utcnow()

        position = html["class"]
        if "reply1" in position:
            self.position = 1
        elif "reply2" in position:
            self.position = 2
        else:
            self.position = 0
        self.children = []

        try:
            links = html.find("div", class_="comment").find_all("a")
            for link in links:
                link.string = link["href"]
            self.content = html.find("div", class_="comment").text
        except AttributeError:
            LOGGER.info(
                "Comment %s by %s has no content, likely embed",
                self.id,
                self.author.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.content = None

        try:
            karma = actions.span.string
            self.karma = int(re.findall(r"[+-]?\d", karma)[0].replace(",", ""))
            self.upvote = join(actions.find_all("a")[1]["href"])
            self.downvote = join(actions.find_all("a")[2]["href"])
            self.approved = True
        except AttributeError:
            self.karma = 0
            self.upvote = None
            self.downvote = None
            self.approved = False
        except IndexError:
            self.downvote = None
            self.approved = True

        self.developer = bool(html.find("span", class_="developer"))
        self.staff = bool(html.find("span", class_="staff"))
        self.subscriber = bool(html.find("span", class_="subscriber"))
        self.guest = self.author.name.lower() == "guest"

        self.embeds = [x["src"] for x in html.find_all("iframe")]

        self.location = html.find("a", class_="related")
        if self.location is not None:
            url = join(self.location["href"])
            page_type = get_page_type(url)
            self.location = Thumbnail(name=self.location.string, url=url, type=page_type)

        try:
            self._hash = html.find("a", title=("Delete", "Undelete"))["href"].split("=")[-1]
        except TypeError:
            self._hash = None

    def is_stale(self):
        """Comments are very volatile. If they are pushed onto another page by other comments
        it becomes impossible to use objects with the previous page number. In addition,
        calculating the new page number is not possible. Pages do not have a defined size but
        rather grow and shrink based on sizes of individual comments. Finally, comments
        also have token that can be used to modify them. These tokens have a hard life of
        30 minutes from the time of the request. This function puts in place several mechanism
        to verify wether or not the object can still be trusted.


        Returns
        --------
        bool
            True, the comment is stale and you should fetch a new version, False you **should**
            be good to continue using it.
        """

        return self._fetch_time + datetime.timedelta(minute=30) > datetime.datetime.utcnow()

    def __repr__(self):
        return (
            f"<Comment author={self.author.name} position={self.position} approved={self.approved}>"
        )


class MissingComment:
    """An object to represent a missing comment. This is used in the cases where a parent comment with
    children is deleted so that the children may still be accessible, missing comment will have the
    same attributes as a :class:`.Comment` but they will all be equal to None or False apart from children
    and the comment position, which will have the children of the comment that was deleted attached to it.
    """

    def __init__(self, position):
        self.id = None
        self.author = None
        self.date = None
        self.position = position
        self.content = None
        self.karma = 0
        self.upvote = None
        self.downvote = None
        self.approved = False
        self.children = []

        self.developer = False
        self.staff = False
        self.subscriber = False
        self.guest = False
        self.embeds = []
        self.location = None

    def __repr__(self):
        return f"<MissingComment position={self.position}>"

    def is_stale(self):
        return True


class MemberProfile:
    """Member profiles are separate entities because they share nothing with the other profile boxes. Where as all
    other profile boxes share at least 4 attributes, a member shares none.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    name : str
        Name of the member
    level : int
        Current level
    progress : float
        Percentage progress to next level
    title : str
        Member title
    avatar : str
        Url of the member avatar
    online : bool
        Whether or not the member is currently online
    last_online :  datetime.datetime
        None if the member is currently online, datetime the user was last seen online
    gender : str
        Gender of the member, can be None
    homepage : str
        URL of the member's homepage
    country : str
        The member's chosen country
    follow : str
        Link to follow a member
    """

    def __init__(self, html: BeautifulSoup):
        profile_raw = html.find("span", string="Profile").parent.parent.parent.find(
            "div", class_="table tablemenu"
        )
        level_raw = profile_raw.find("h5", string="Level").parent.span.div
        self.name = html.find("meta", property="og:title")["content"]

        self.level = int(level_raw.find("span", class_="level").string)
        self.progress = float(
            "0." + level_raw.find("span", class_="info").strong.string.replace("%", "")
        )
        self.title = level_raw.find("span", class_="info").a.string

        self.avatar = profile_raw.find("div", class_="avatarinfo").img["src"]
        self.online = bool(profile_raw.find("h5", string="Status"))
        last_online = profile_raw.find("h5", string="Last Online")
        self.last_online = (
            get_date(last_online.parent.span.time["datetime"]) if last_online else None
        )

        try:
            self.gender = profile_raw.find("h5", string="Gender").parent.span.string.strip()
        except AttributeError:
            LOGGER.info(
                "Member %s has not publicized their gender",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.gender = None

        try:
            self.homepage = html.find("h5", string="Homepage").parent.span.a["href"]
        except AttributeError:
            self.homepage = None
            LOGGER.info(
                "Member %s has no homepage", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )

        try:
            self.country = profile_raw.find("h5", string="Country").parent.span.string.strip()
        except AttributeError:
            self.country = None
            LOGGER.info(
                "Member %s country is not visible (happens when not logged in)",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        try:
            self.follow = join(html.find("a", title="Follow")["href"])
        except TypeError:
            LOGGER.info(
                "Can't watch yourself, narcissist...", exc_info=LOGGER.level >= logging.DEBUG
            )
            self.follow = None

    def __repr__(self):
        return f"<MemberProfile name={self.name}>"


class MemberStatistics:
    """Similarly, a member statistics shared no common ground with other stats and therefore there was a
    need for a separate object.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    watchers : int
        How many members are following this member
    acivity_points : int
        Activity points
    comments : int
        How many comments the member has made
    tags : int
        How many tags the member has created
    visits : int
        How many people have viewed this page
    site_visits : int
        How many time this user has visited the site
    today : int
        How many people have viewed this page today
    time : int
        How many seconds the member has spent online
    rank : int
        The member's current rank (compared to other members)
    total : int
        the maximum rank
    """

    def __init__(self, html: BeautifulSoup):
        def get(parent):
            return parent.a.string.strip() if parent.a else parent.span.string.strip()

        name = html.find("meta", property="og:title")["content"]
        misc = html.find_all(
            "h5",
            string=("Watchers", "Activity Points", "Comments", "Tags", "Site visits"),
        )
        self.__dict__.update(
            {
                stat.string.lower().replace(" ", "_"): int(normalize(get(stat.parent)))
                for stat in misc
            }
        )

        visits = normalize(html.find("h5", string="Visitors").parent.a.string)
        self.visits, self.today = get_views(visits)

        time, mapping = html.find("h5", string="Time Online").parent.span.string.strip().split(" ")
        self.time = time_mapping[mapping.replace("s", "")] * int(time)

        try:
            rank = normalize(html.find("h5", string="Rank").parent.span.string).split("of")
            self.rank = int(rank[0].replace(",", ""))
            self.total = int(rank[1].replace(",", ""))
        except AttributeError:
            self.rank = 0
            self.total = 0
            LOGGER.info("Member %s has no rank", name, exc_info=LOGGER.level >= logging.DEBUG)

    def __repr__(self):
        return f"<MemberStatistics rank={self.rank}/{self.total}>"


class PlatformStatistics:
    """Stats for platform pages.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    hardware : int
        Number of harware created for this platform
    software : int
        Number of software created for this platform
    engines : int
        Number of engines created for this platform
    games : int
        Number of games created for this platform
    mods : int
        Number of mods created for this platform
    """

    def __init__(self, html: BeautifulSoup):
        headings = ("Hardware", "Software", "Engines", "Games", "Mods")
        html_headings = html.find_all("h5", string=headings)
        self.__dict__.update(
            {
                headings[html_headings.index(x)].lower(): int(normalize(x.parent.span.a.string))
                for x in html_headings
            }
        )

    def __repr__(self):
        return "<PlatformStatistics>"


class PartialArticle:
    """A partial article is an article object missing attributes due to being parsed from the front page
    intead of from the article page itself. In general, it' is simple enough for previewing the article
    but if you need a full article with comments, profile, ect... Then parse it with the method

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    name : str
        Name of the articles
    url : str
        Link to the article
    date : datetime.datetime
        Date the article was published
    type : ArticleCategory
        Type of the article
    content : str
        html of the article content
    plaintext : str
        plaintext of the article content (without html)
    """

    def __init__(self, html: BeautifulSoup):
        meta_raw = html.find("div", class_="row rowcontent rownoimage clear")

        self.name = meta_raw.h4.a.string
        self.url = join(meta_raw.h4.a["href"])
        self.date = get_date(meta_raw.find("time")["datetime"])
        try:
            self.type = ArticleCategory[
                meta_raw.find("span", class_="subheading").text.strip().split(" ")[0].lower()
            ]
        except KeyError:
            self.type = ArticleCategory.news

        content = html.find("div", class_="row rowcontent rowcontentnext clear")
        self.content = str(content)
        self.plaintext = content.text

    def __repr__(self):
        return f"<PartialArticle title={self.name}>"

    def get_article(self) -> Article:
        """Returns the full article object of this article.

        Returns
        --------
        Article
            The complete article object
        """
        from .pages.article import Article

        return Article(get_page(self.url))


class Option:
    """Represents one of the choice from the poll they are attached to, should not be created
    manually, prefer relying on the Poll.

    Attributes
    -----------
    id : int
        The id of the option, can be None and will be None in most cases.
    text : str
        The option's text
    votes : int
        The number of votes that have been cast on this option
    percent : int
        The percent of all votes that have been cast on this option
    """

    def __init__(self, **kwargs):
        self.id = kwargs.get("id", None)
        self.text = kwargs.get("text")
        self.votes = kwargs.get("votes")
        self.percent = kwargs.get("percent")

    def __repr__(self):
        return f"<Option text={self.text}>"


T = TypeVar("T")


class ModDBList(collections.abc.MutableSequence[T], Generic[T]):
    """Base List type for the lib

    Attributes
    -----------
    current_page : int
        The page of results this objects represents
    total_pages : int
        The total amount of result pages available
    total_results : int
        The total amount of results available
    """

    def __init__(self, **kwargs):
        self._results: List[T] = kwargs.pop("results")
        self._params: dict = kwargs.pop("params", {})
        self._url = kwargs.pop("url")
        self.total_pages = kwargs.pop("total_pages")
        self.current_page = kwargs.pop("current_page")
        self.total_results = kwargs.pop("total_results")

    def _parse_method(self, html: BeautifulSoup):
        raise NotImplementedError

    def _do_request(self, **kwargs):
        page = kwargs.pop("page", self.current_page)
        params = {**self._params, **kwargs}

        html = get_page(f"{self._url}/page/{page}", params=params)
        results, current_page, total_pages, total_results = self._parse_method(html)

        return self.__class__(
            results=results,
            params=params,
            url=self._url,
            total_pages=total_pages,
            current_page=current_page,
            total_results=total_results,
        )

    def next_page(self) -> Self:
        """Returns the next page of results as either a CommentList if you are retriving comments or
        as a ResultList if it's literally anything else.

        Returns
        --------
        Self
            The new search objects containing a new set of results.

        Raises
        -------
        ValueError
            There is no next page
        """
        if self.current_page == self.total_pages:
            raise ValueError("Reached last page already")

        return self.to_page(self.current_page + 1)

    def previous_page(self) -> Self:
        """Returns the previous page of results as either a CommentList if you are retriving comments or
        as a ResultList if it's literally anything else.

        Returns
        --------
        Self
            The new list-like object of results.

        Raises
        -------
        ValueError
            There is no previous page
        """
        if self.current_page == 1:
            raise ValueError("Reached first page already")

        return self.to_page(self.current_page - 1)

    def to_page(self, page: int) -> Self:
        """Returns the desired page of results as either a CommentList if you are retriving comments or
        as a ResultList if it's literally anything else.

        Parameters
        -----------
        page : int
            A page number within the range 1 - max_page inclusive

        Returns
        --------
        Self
            The new list-like object of results.

        Raises
        -------
        ValueError
            This page does not exist
        """
        if page < 1 or page > self.total_pages:
            raise ValueError(f"Please pick a page between 1 and {self.total_pages}")

        return self._do_request(page=page)

    def get_all_results(self) -> Self:
        """An expensive methods that iterates over every page of the result query and returns all
        the results. This may return more results than you expected if new page have fit the criteria
        while iterating.

        Returns
        --------
        Union[CommentList[Any], ResultList[Any]]
            The list of things you were searching for
        """
        search = self.to_page(1)
        results = list(search)

        while True:
            try:
                search = search.next_page()
            except ValueError:
                break
            else:
                results.extend(search)
                LOGGER.info("Parsed page %s/%s", search.current_page, search.total_pages)

        def key(element):
            if isinstance(element, Comment):
                return element.id
            else:
                return element.name

        search._results = list({key(e): e for e in results}.values())
        return search

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} pages={self.current_page}/{self.total_pages}, results={self._results}>"

    def __getitem__(self, index: int) -> T:
        return self._results.__getitem__(index)

    def __delitem__(self, index: int):
        self._results.__delitem__(index)

    def __len__(self) -> int:
        return self._results.__len__()

    def __setitem__(self, key: int, value: T):
        self._results.__setitem__(key, value)

    def insert(self, index: int, value: T):
        self._results.insert(index, value)


class ResultList(ModDBList[Thumbnail]):
    """Represents a list of result gotten from one of the many get methods the library uses. This is returned
    over a regular list because it has additional methods that allow for easily go through all the results. In
    the same way that the moddb site works, you don't have to re-run the query manually to get the next page,
    you simply click a button, same here, you don't have to recall the base get method, simply use on of the
    methods here to traverse the results. This emulates a list and will behave like one, so you
    can use any of the regular list operators in addition to the methods defined below

    Attributes
    -----------
    current_page : int
        The page of results this objects represents
    total_pages : int
        The total amount of result pages available
    total_results : int
        The total amount of results available
    """

    def _parse_method(self, html: BeautifulSoup):
        return _parse_results(html)

    def resort(self, new_sort: Tuple[str, str]) -> Self:
        """Allows you to sort the whole search by a new sorting parameters. Returns a new search object.

        Parameters
        -----------
        new_sort : Tuple[str, str]
            The new sorting tuple to check by

        Returns
        -------
        ResultList
            The new set of results with the updated sort order
        """
        return self._do_request(sort=f"{new_sort[0]}-{new_sort[1]}")

    def __contains__(self, element: Thumbnail) -> bool:
        return get(self._results, name=element.name) is not None


class CommentList(ModDBList[Comment]):
    """Represents a list of comments. This emulates a list and will behave like one, so you
    can use any of the regular list operators in addition to the methods defined below.

    Attributes
    -----------
    current_page : int
        The page of results this objects represents
    total_pages : int
        The total amount of result pages available
    total_results : int
        The total amount of results available
    """

    def _parse_method(self, html: BeautifulSoup):
        return _parse_comments(html)

    def __contains__(self, element: Comment) -> bool:
        return get(self._results, name=element.name) is not None

    def flatten(self) -> List[Comment]:
        """Returns a 'flattened' list of comments where children of comments are added right
        after the parent comment so:

        [ Comment1 ]
            ├── Comment2\n
            |   ├── Comment3\n
            |   └── Comment4\n
            └── Comment5

        would become:

        [Comment1, Comment2, Comment3, Comment4, Comment5]

        Returns
        --------
        List[Comment]
            The flattened list of comments
        """
        top_list = []
        for comment in self._results:
            top_list.append(comment)
            for child in comment.children:
                top_list.append(child)
                top_list.extend(child.children)

        return top_list


class Mirror:
    """Represents a download mirror from which the user can download a file

    Attributes
    -----------
    name : str
        The name of the mirror
    index : int
        The index of the mirror, as multiple mirrors
        have the same name. Index starts at 1
    city : str
        Alpha 2 code, or full name, of the city the server is located
        in. Sometimes represents a country.
    country : str
        Alpha 2 code for the country the server is
        located in. Sometimes represents a continent.
    served : int
        How many downloads of this file this mirror has
        served
    capacity : float
        The current capacity of this server as a percentage.
        E.g. 35.5 -> 35.5%. Lower is better for speed.
    """

    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.index = kwargs.get("index")
        self.city = kwargs.get("city")
        self.country = kwargs.get("country")
        self.served = kwargs.get("served")
        self.capacity = kwargs.get("capacity")
        self._url = kwargs.get("url")

    def __repr__(self):
        return f"<Mirror name={self.name} index={self.index} >"


PartialTag = collections.namedtuple("PartialTag", "name name_id url")


class Tag:
    """Represents a tag, useful to vote on stuff

    Parameters
    -----------
    id : Optional[int]
        ID of the tag. None if obtained from parsing a page
    name_id : str
        Name id of the tag
    name : str
        Name of the tag
    date : Optional[datetime.datetime]
        Creation date of the tag. None if obtained from parsing a page
    official : Optional[bool]
        Whether the tag is official or user created. None if obtained from parsing a page
    sitearea : int
        Site area
    siteareaid : int
        Site area id
    positive : int
        Number of positive votes
    negative : int
        Number of negative votes
    rank : Optiona[int]
        Rank of the tag in trending. None if obtained from parsing a page
    url : str
        Url to the tag
    """

    def __init__(self, **kwargs):
        self.id = int(kwargs.pop("id"))
        self.date = datetime.datetime.fromtimestamp(int(kwargs.pop("date")))
        self.official = kwargs.pop("official") == "1"
        self.sitearea = int(kwargs.pop("sitearea"))
        self.siteareaid = int(kwargs.pop("siteareaid"))

        self.positive = int(kwargs.pop("positive"))
        self.negative = int(kwargs.pop("negative"))

        self.name_id = kwargs.pop("tagid")
        self.name = kwargs.pop("tag")
        self.rank = int(kwargs.pop("trending"))
        self.url = f"{BASE_URL}/tags/{self.name_id}"

    def __repr__(self) -> str:
        return f"< Tag id={self.id} name_id={self.name_id} >"

    def _get_members(self):
        """Get a list of the members that have voted for this tag

        Returns
        ---------
        List[Thumbnail]
            List of member typed thumbnail
        """
        LOGGER.warning("_get_members is undocumented and unreliable for the time being")
        params = {
            "ajax": "t",
            "tag": self.name_id,
            "sitearea": get_siteareaid(self.sitearea),
            "siteareaid": self.siteareaid,
            "hash": generate_hash(),
        }

        resp = get_page(f"{BASE_URL}/tags/ajax/who", params=params)

        return [
            Thumbnail(url=join(thumb["href"]), name=thumb.string, type=ThumbnailType.member)
            for thumb in resp.find("div", class_="successboxachtung").find_all("a")
        ]



import re

import bs4

from ..boxes import Thumbnail
from ..enums import ThumbnailType
from ..utils import LOGGER, get_page, get_page_type, join
from . import opinion


class FrontPage:
    """An object representing the front page of  More of less just a long suggestion of the hottest mods,
    games, articles and files of the moment

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Attributes
    -----------
    slider : List[Thumnail]
        A list of object like thumbnails presented by the slider present on
        the front page, this is a catered list of promoted content, can contain
        any type of object.
    articles : List[Thumbnail]
        A list of article like thumbnail objects representing the suggested
        articles on the front page.
    mods : List[Thumbnail]
        A list of mod like thumbnail objects representing the suggested
        mods on the front page.
    games : List[Thumbnail]
        A list of game like thumbnail objects representing the suggested
        games on the front page.
    files : List[Thumbnail]
        A list of file like thumbnail objects representing the suggested
        files on the front page.
    poll_url : str
        The url to the poll on the front page
    """

    def __init__(self, html: bs4.BeautifulSoup):
        slider = html.find("div", class_="rotatorholder").find_all("div", class_="rotatorbox")
        self.slider = []
        for x in slider:
            name = x.a.find("h2")
            summary = x.a.find("p")

            image = None
            if "style" in x.div:
                re_search = re.search(r"\((.*)\)", x.div["style"])
                if re_search:
                    image = re_search.group(1)
            elif "data-bg" in x.div:
                image = x.div["data-bg"]

            try:
                page_type = get_page_type(x.a["href"])
            except IndexError:
                LOGGER.warning(
                    "Unabled to get page type from %s for front page, skipping", x.a["href"]
                )
                continue

            thumbnail = Thumbnail(
                name=name.string if name else None,
                url=x.a["href"],
                summary=summary.string if summary else None,
                image=image,
                type=page_type,
            )

            self.slider.append(thumbnail)

        articles = (
            html.find("span", string="Latest Articles")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[:-1]
        )
        self.articles = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.article,
                image=x.a.img["src"],
                summary=x.find("p").string,
                date=x.find("time")["datetime"],
            )
            for x in articles
        ]

        mods = (
            html.find("span", string="Popular Mods")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[1:]
        )
        self.mods = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.mod,
                image=x.a.img["src"],
            )
            for x in mods
        ]

        games = (
            html.find("span", string="Popular Games")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[1:]
        )
        self.games = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.game,
                image=x.a.img["src"],
            )
            for x in games
        ]

        jobs = (
            html.find("span", string="Jobs")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[1:]
        )
        self.jobs = [
            Thumbnail(
                name=x.find("a").string,
                url=x.find("a")["href"],
                type=ThumbnailType.job,
            )
            for x in jobs
        ]

        files = (
            html.find("span", string="Popular Files")
            .parent.parent.parent.find("div", class_="table")
            .find_all("div", recursive=False)[1:]
        )
        self.files = [
            Thumbnail(
                name=x.a["title"],
                url=x.a["href"],
                type=ThumbnailType.file,
                image=x.a.img["src"],
            )
            for x in files
        ]

        try:
            # maybe they haven't voted yet and the vote bar is there
            self.poll_url = html.find("div", class_="poll").form["action"]
        except AttributeError:
            # maybe they've already voted and the result link is there
            self.poll_url = join(
                html.find("div", class_="poll").find("a", class_="results")["href"]
            )

        self._html = html
        self._poll = None

    def __repr__(self):
        return f"<FrontPage articles={len(self.articles)} mods={len(self.mods)} games={len(self.games)} files={len(self.files)}>"

    def get_poll(self) -> opinion.Poll:
        """Get the full item of the front page poll. This
        result is cached after the first call.

        Returns
        --------
        Poll
            The returned poll
        """
        if self._poll is None:
            self._poll = opinion.Poll(get_page(self.poll_url))

        return self._poll



import bs4

from ..enums import SearchCategory
from ..utils import concat_docs
from .base import PageMetaClass
from .mixins import GetAddonsMixin, GetModsMixin


@concat_docs
class Game(PageMetaClass, GetModsMixin, GetAddonsMixin):
    """A subclass of Page plus a method to get all the mods.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.


    Filtering
    ----------
    released : :class:`.Status`
        The status of the game (released, unreleased)
    genre : :class:`.Genre`
        The genre of the game (fps, tower defense)
    theme : :class:`.Theme`
        The theme of the game (fantasy, action)
    indie : :class:`.Scope`
        Whether the game is triple AAA or indie
    players : :class:`.PlayerStyle`
        Player styles of the game (co-op, singleplayer)
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first
    """

    def __init__(self, html: bs4.BeautifulSoup):
        super().__init__(html, SearchCategory.games)
